import sys
import csv
import io
from pathlib import Path
from itertools import chain
from decimal import Decimal
from typing import NamedTuple, Iterator, List, Dict, Any
from datetime import datetime, timezone

from more_itertools import unique_everseen

from .models import State


class Solve(NamedTuple):
    puzzle: str
    category: str
    scramble: str
    time: Decimal
    penalty: Decimal
    dnf: bool
    when: datetime
    comment: str

    def to_csv_list(self) -> List[str]:
        penalty_code = 0
        if self.penalty == Decimal("2"):
            penalty_code = 1
        if self.dnf:
            penalty_code = 2
        return [
            self.puzzle,
            self.category,
            str(int(self.time * 1000)),
            str(int(self.when.timestamp() * 1000)),
            self.scramble,
            str(penalty_code),
            self.comment,
        ]

    @property
    def _prompt_defaults(self) -> Dict[str, Any]:
        return {
            "transformed_puzzle": self.puzzle,
            "transformed_event_description": self.category,
        }

    def _transform_map(self) -> Dict[str, Any]:
        return dict(
            state=State.DNF if self.dnf else State.SOLVED,
            scramble=self.scramble,
            comment=self.comment,
            time=self.time,
            penalty=self.penalty,
            when=self.when,
            full_time=self.time + self.penalty,
        )


HEADER: str = "Puzzle,Category,Time(millis),Date(millis),Scramble,Penalty,Comment"


def serialize_solves(solves: List[Solve]) -> str:
    buf = io.StringIO()
    buf.write(HEADER)
    buf.write("\n")
    buf.flush()
    writer = csv.writer(buf, delimiter=";", quoting=csv.QUOTE_ALL)
    writer.writerows([r.to_csv_list() for r in solves])
    return str(buf.getvalue())


def parse_file(path: Path) -> Iterator[Solve]:
    with path.open("r", newline="") as f:
        reader = csv.reader(f, delimiter=";")
        next(reader)
        for row in reader:
            try:
                [puzzle, category, time, date, scramble, penalty, comment] = row
            except ValueError:
                print(
                    f"Could not parse line, expected 7 fields, found {len(row)}: {row}",
                    file=sys.stderr,
                )
                raise
            upenalty = 0
            is_dnf = penalty == "2"
            if penalty == "1":
                upenalty = 2
            yield Solve(
                puzzle=puzzle,
                category=category,
                scramble=scramble,
                time=Decimal(time) / 1000,
                dnf=is_dnf,
                penalty=Decimal(upenalty),
                when=datetime.fromtimestamp(int(date) / 1000, tz=timezone.utc),
                comment=comment,
            )


def parse_files(paths: List[Path]) -> Iterator[Solve]:
    yield from unique_everseen(chain(*(parse_file(p) for p in paths)))



from typing import NamedTuple, Optional
from functools import lru_cache


class CSTimerScramble(NamedTuple):
    scramble_code: str
    category: Optional[str]
    name: Optional[str]


# instead of dict/mapping just save all possible data
# there is some overlap between what the 'correct' description
# to return for a scramble is depending on which category the user
# had selected, so not going to try to be optimal here
SCRAMBLES = [
    CSTimerScramble(scramble_code="333", category=None, name="3x3x3"),
    CSTimerScramble(scramble_code="222so", category=None, name="2x2x2"),
    CSTimerScramble(scramble_code="444wca", category=None, name="4x4x4"),
    CSTimerScramble(scramble_code="555wca", category=None, name="5x5x5"),
    CSTimerScramble(scramble_code="666wca", category=None, name="6x6x6"),
    CSTimerScramble(scramble_code="777wca", category=None, name="7x7x7"),
    CSTimerScramble(scramble_code="333ni", category=None, name="3x3 bld"),
    CSTimerScramble(scramble_code="333fm", category=None, name="3x3 fm"),
    CSTimerScramble(scramble_code="333oh", category=None, name="3x3 oh"),
    CSTimerScramble(scramble_code="clkwca", category=None, name="clock"),
    CSTimerScramble(scramble_code="mgmp", category=None, name="megaminx"),
    CSTimerScramble(scramble_code="pyrso", category=None, name="pyraminx"),
    CSTimerScramble(scramble_code="skbso", category=None, name="skewb"),
    CSTimerScramble(scramble_code="sqrs", category=None, name="sq1"),
    CSTimerScramble(scramble_code="444bld", category=None, name="4x4 bld"),
    CSTimerScramble(scramble_code="555bld", category=None, name="5x5 bld"),
    CSTimerScramble(scramble_code="r3ni", category=None, name="3x3 mbld"),
    CSTimerScramble(scramble_code="input", category=None, name="Extern"),
    CSTimerScramble(scramble_code="remoteComp", category=None, name="Competition"),
    CSTimerScramble(scramble_code="remoteOther", category=None, name="Remote"),
    CSTimerScramble(scramble_code="333", category="wca", name="random state (WCA)"),
    CSTimerScramble(scramble_code="333o", category="wca", name="random move"),
    CSTimerScramble(scramble_code="333noob", category="wca", name="3x3x3 for noobs"),
    CSTimerScramble(scramble_code="edges", category="wca", name="edges only"),
    CSTimerScramble(scramble_code="corners", category="wca", name="corners only"),
    CSTimerScramble(scramble_code="ll", category="wca", name="last layer"),
    CSTimerScramble(scramble_code="zbll", category="wca", name="zb last layer"),
    CSTimerScramble(scramble_code="cll", category="wca", name="corners of last layer"),
    CSTimerScramble(scramble_code="ell", category="wca", name="edges of last layer"),
    CSTimerScramble(scramble_code="lse", category="wca", name="last six edges"),
    CSTimerScramble(
        scramble_code="lsemu", category="wca", name="last six edges&ltM,U&gt"
    ),
    CSTimerScramble(scramble_code="cmll", category="wca", name="Roux L10P"),
    CSTimerScramble(scramble_code="f2l", category="wca", name="cross solved"),
    CSTimerScramble(
        scramble_code="lsll2", category="wca", name="last slot + last layer"
    ),
    CSTimerScramble(scramble_code="2gll", category="wca", name="2GLL"),
    CSTimerScramble(scramble_code="zbls", category="wca", name="ZBLS"),
    CSTimerScramble(scramble_code="zzll", category="wca", name="ZZLL"),
    CSTimerScramble(scramble_code="oll", category="wca", name="OLL"),
    CSTimerScramble(scramble_code="pll", category="wca", name="PLL"),
    CSTimerScramble(scramble_code="eoline", category="wca", name="EOLine"),
    CSTimerScramble(scramble_code="easyc", category="wca", name="easy cross"),
    CSTimerScramble(scramble_code="333ft", category="wca", name="3x3 ft"),
    CSTimerScramble(scramble_code="222so", category="wca", name="random state (WCA)"),
    CSTimerScramble(scramble_code="222o", category="wca", name="optimal"),
    CSTimerScramble(scramble_code="2223", category="wca", name="3-gen"),
    CSTimerScramble(scramble_code="222eg", category="wca", name="EG"),
    CSTimerScramble(scramble_code="222eg0", category="wca", name="EG0"),
    CSTimerScramble(scramble_code="222eg1", category="wca", name="EG1"),
    CSTimerScramble(scramble_code="222eg2", category="wca", name="EG2"),
    CSTimerScramble(scramble_code="222nb", category="wca", name="No Bar"),
    CSTimerScramble(scramble_code="444wca", category="wca", name="WCA"),
    CSTimerScramble(scramble_code="444m", category="wca", name="random move"),
    CSTimerScramble(scramble_code="444", category="wca", name="SiGN"),
    CSTimerScramble(scramble_code="444yj", category="wca", name="YJ"),
    CSTimerScramble(scramble_code="4edge", category="wca", name="4x4x4 edges"),
    CSTimerScramble(scramble_code="RrUu", category="wca", name="R,r,U,u"),
    CSTimerScramble(scramble_code="555wca", category="wca", name="WCA"),
    CSTimerScramble(scramble_code="555", category="wca", name="SiGN"),
    CSTimerScramble(scramble_code="5edge", category="wca", name="5x5x5 edges"),
    CSTimerScramble(scramble_code="666wca", category="wca", name="WCA"),
    CSTimerScramble(scramble_code="666si", category="wca", name="SiGN"),
    CSTimerScramble(scramble_code="666p", category="wca", name="prefix"),
    CSTimerScramble(scramble_code="666s", category="wca", name="suffix"),
    CSTimerScramble(scramble_code="6edge", category="wca", name="6x6x6 edges"),
    CSTimerScramble(scramble_code="777wca", category="wca", name="WCA"),
    CSTimerScramble(scramble_code="777si", category="wca", name="SiGN"),
    CSTimerScramble(scramble_code="777p", category="wca", name="prefix"),
    CSTimerScramble(scramble_code="777s", category="wca", name="suffix"),
    CSTimerScramble(scramble_code="7edge", category="wca", name="7x7x7 edges"),
    CSTimerScramble(scramble_code="clk", category="wca", name="jaap"),
    CSTimerScramble(scramble_code="clkwca", category="wca", name="wca"),
    CSTimerScramble(scramble_code="clko", category="wca", name="optimal"),
    CSTimerScramble(scramble_code="clkc", category="wca", name="concise"),
    CSTimerScramble(scramble_code="clke", category="wca", name="efficient pin order"),
    CSTimerScramble(scramble_code="mgmp", category="wca", name="WCA"),
    CSTimerScramble(scramble_code="mgmc", category="wca", name="Carrot"),
    CSTimerScramble(scramble_code="mgmo", category="wca", name="old style"),
    CSTimerScramble(scramble_code="pyrso", category="wca", name="random state (WCA)"),
    CSTimerScramble(scramble_code="pyro", category="wca", name="optimal"),
    CSTimerScramble(scramble_code="pyrm", category="wca", name="random move"),
    CSTimerScramble(scramble_code="pyrl4e", category="wca", name="L4E"),
    CSTimerScramble(scramble_code="pyr4c", category="wca", name="4 tips"),
    CSTimerScramble(scramble_code="pyrnb", category="wca", name="No bar"),
    CSTimerScramble(scramble_code="skbso", category="wca", name="random state (WCA)"),
    CSTimerScramble(scramble_code="skbo", category="wca", name="optimal"),
    CSTimerScramble(scramble_code="skb", category="wca", name="random move"),
    CSTimerScramble(scramble_code="skbnb", category="wca", name="No bar"),
    CSTimerScramble(scramble_code="sqrs", category="wca", name="random state (WCA)"),
    CSTimerScramble(scramble_code="sqrcsp", category="wca", name="CSP"),
    CSTimerScramble(scramble_code="sq1h", category="wca", name="face turn metric"),
    CSTimerScramble(scramble_code="sq1t", category="wca", name="twist metric"),
    CSTimerScramble(scramble_code="15prp", category="other", name="random state URLD"),
    CSTimerScramble(scramble_code="15prap", category="other", name="random state ^<>v"),
    CSTimerScramble(
        scramble_code="15prmp", category="other", name="random state Blank"
    ),
    CSTimerScramble(scramble_code="15p", category="other", name="random move URLD"),
    CSTimerScramble(scramble_code="15pat", category="other", name="random move ^<>v"),
    CSTimerScramble(scramble_code="15pm", category="other", name="random move Blank"),
    CSTimerScramble(scramble_code="8prp", category="other", name="random state URLD"),
    CSTimerScramble(scramble_code="8prap", category="other", name="random state ^<>v"),
    CSTimerScramble(scramble_code="8prmp", category="other", name="random state Blank"),
    CSTimerScramble(scramble_code="133", category="other", name="1x3x3 (Floppy Cube)"),
    CSTimerScramble(scramble_code="223", category="other", name="2x2x3 (Tower Cube)"),
    CSTimerScramble(scramble_code="233", category="other", name="2x3x3 (Domino)"),
    CSTimerScramble(scramble_code="334", category="other", name="3x3x4"),
    CSTimerScramble(scramble_code="335", category="other", name="3x3x5"),
    CSTimerScramble(scramble_code="336", category="other", name="3x3x6"),
    CSTimerScramble(scramble_code="337", category="other", name="3x3x7"),
    CSTimerScramble(scramble_code="888", category="other", name="8x8x8"),
    CSTimerScramble(scramble_code="999", category="other", name="9x9x9"),
    CSTimerScramble(scramble_code="101010", category="other", name="10x10x10"),
    CSTimerScramble(scramble_code="111111", category="other", name="11x11x11"),
    CSTimerScramble(scramble_code="cubennn", category="other", name="NxNxN"),
    CSTimerScramble(scramble_code="gearso", category="other", name="random state"),
    CSTimerScramble(scramble_code="gearo", category="other", name="optimal"),
    CSTimerScramble(scramble_code="gear", category="other", name="random move"),
    CSTimerScramble(scramble_code="cm3", category="other", name=None),
    CSTimerScramble(scramble_code="cm2", category="other", name=None),
    CSTimerScramble(scramble_code="giga", category="other", name="Pochmann"),
    CSTimerScramble(scramble_code="heli", category="other", name=None),
    CSTimerScramble(scramble_code="redim", category="other", name="MoYu"),
    CSTimerScramble(scramble_code="redi", category="other", name="old"),
    CSTimerScramble(scramble_code="ivyso", category="other", name="random state"),
    CSTimerScramble(scramble_code="ivyo", category="other", name="optimal"),
    CSTimerScramble(scramble_code="ivy", category="other", name="random move"),
    CSTimerScramble(scramble_code="mpyr", category="other", name=None),
    CSTimerScramble(scramble_code="prcp", category="other", name="Pochmann"),
    CSTimerScramble(scramble_code="prco", category="other", name="old style"),
    CSTimerScramble(scramble_code="sia113", category="other", name="1x1x3 block"),
    CSTimerScramble(scramble_code="sia123", category="other", name="1x2x3 block"),
    CSTimerScramble(scramble_code="sia222", category="other", name="2x2x2 block"),
    CSTimerScramble(scramble_code="sq2", category="other", name=None),
    CSTimerScramble(scramble_code="sfl", category="other", name=None),
    CSTimerScramble(scramble_code="ssq1t", category="other", name="twist metric"),
    CSTimerScramble(scramble_code="ufo", category="other", name="Jaap style"),
    CSTimerScramble(
        scramble_code="fto", category="other", name="FTO (Face-Turning Octahedron)"
    ),
    CSTimerScramble(scramble_code="2gen", category="special", name="2-generator R,U"),
    CSTimerScramble(scramble_code="2genl", category="special", name="2-generator L,U"),
    CSTimerScramble(
        scramble_code="roux", category="special", name="Roux-generator M,U"
    ),
    CSTimerScramble(
        scramble_code="3gen_F", category="special", name="3-generator F,R,U"
    ),
    CSTimerScramble(
        scramble_code="3gen_L", category="special", name="3-generator R,U,L"
    ),
    CSTimerScramble(scramble_code="RrU", category="special", name="3-generator R,r,U"),
    CSTimerScramble(scramble_code="half", category="special", name="half turns only"),
    CSTimerScramble(
        scramble_code="lsll", category="special", name="last slot + last layer (old)"
    ),
    CSTimerScramble(scramble_code="bic", category="special", name="Bicube"),
    CSTimerScramble(scramble_code="bsq", category="special", name="Square-1 /,(1,0)"),
    CSTimerScramble(scramble_code="minx2g", category="special", name="2-generator R,U"),
    CSTimerScramble(
        scramble_code="mlsll", category="special", name="last slot + last layer"
    ),
    CSTimerScramble(scramble_code="r3", category="special", name="lots of 3x3x3s"),
    CSTimerScramble(scramble_code="r234", category="special", name="234 relay"),
    CSTimerScramble(scramble_code="r2345", category="special", name="2345 relay"),
    CSTimerScramble(scramble_code="r23456", category="special", name="23456 relay"),
    CSTimerScramble(scramble_code="r234567", category="special", name="234567 relay"),
    CSTimerScramble(scramble_code="r234w", category="special", name="234 relay (WCA)"),
    CSTimerScramble(
        scramble_code="r2345w", category="special", name="2345 relay (WCA)"
    ),
    CSTimerScramble(
        scramble_code="r23456w", category="special", name="23456 relay (WCA)"
    ),
    CSTimerScramble(
        scramble_code="r234567w", category="special", name="234567 relay (WCA)"
    ),
    CSTimerScramble(scramble_code="111", category="jokes", name="x y z"),
    CSTimerScramble(scramble_code="-1", category="jokes", name=None),
    CSTimerScramble(scramble_code="112", category="jokes", name=None),
    CSTimerScramble(scramble_code="lol", category="jokes", name=None),
    CSTimerScramble(scramble_code="eide", category="jokes", name=None),
]


@lru_cache(maxsize=None)
def parse_scramble_type(code: str) -> CSTimerScramble:
    for scr in SCRAMBLES:
        if code == scr.scramble_code:
            return scr
    raise KeyError(f"Could not find matching scramble code {code}")



from decimal import Decimal
from statistics import mean, StatisticsError
from typing import NamedTuple, Optional, List, Tuple, Dict, Union
from math import inf

from .models import State, Operation, Solve
from .error import Res
from .timeformat import format_decimal


def solves_to_float(solves_dec: List[Solve]) -> List[float]:
    return [inf if s.state != State.SOLVED else float(s.full_time) for s in solves_dec]


def findminmax(solves: Union[List[Solve], List[float]]) -> Tuple[int, int]:
    """
    returns indexes of min, max
    """
    solves_enumd: List[Tuple[int, float]]
    if isinstance(solves[0], Solve):
        solves_enumd = list(enumerate(solves_to_float(solves)))  # type: ignore[arg-type]
    else:
        solves_enumd = list(enumerate(solves))  # type: ignore[arg-type]
    min_i = min(solves_enumd, key=lambda s: s[1])[0]
    max_i = max(solves_enumd, key=lambda s: s[1])[0]
    return min_i, max_i


def operation_code(operation: Operation, count: int, solves_len: int) -> str:
    if operation == "global_mean":
        return f"GlobalMean ({count}/{solves_len})"
    else:
        if operation == "mean":
            return f"Mo{count}"
        else:
            return f"Ao{count}"


# e.g. average of 5, mean of 5
class Grouping(NamedTuple):
    operation: Operation
    state: State
    result: Decimal
    solves: List[Solve]
    solve_count: Optional[int]

    def __str__(self) -> str:
        return self.describe()

    @property
    def operation_code(self) -> str:
        assert (
            self.solve_count is not None
        ), f"While computing operation code text, count is not set while operation is {self.operation}"
        return operation_code(self.operation, self.solve_count, len(self.solves))

    @property
    def lhs(self) -> str:
        return "DNF" if self.state != State.SOLVED else format_decimal(self.result)

    def describe_average(self) -> str:
        desc = [s.describe() for s in self.solves]
        if self.operation == "average" and self.state == State.SOLVED:
            mini, maxi = findminmax(self.solves)
            # surround min/max with parenthesis
            desc[mini] = f"({desc[mini]})"
            desc[maxi] = f"({desc[maxi]})"
        return f"{self.lhs} = {' '.join(desc)}"

    def describe(self) -> str:
        return f"{self.operation_code}: {self.describe_average()}"


def grouped(
    solves_dc: List[Solve],
    *,
    operation: Operation,
    solves_flt: Optional[List[float]] = None,
    count: Optional[int] = None,
) -> Res[Grouping]:
    """
    solves should be sorted/ordered prior to doing a grouping
    """
    if solves_flt is None:
        solves_flt = solves_to_float(solves_dc)
    # error checking
    if operation == "average" or operation == "mean":
        if count is None:
            count = len(solves_flt)
        else:
            if len(solves_flt) < count:
                return ValueError(
                    f"Only have {len(solves_flt)} solves, cannot compute {operation} of {count}"
                )
    if operation == "average" and (
        len(solves_flt) < 3 or (count is not None and count < 3)
    ):
        return ValueError("Cannot do operation 'average' with less than 3 solves")

    # take first 'count' elements if user passed a larger list
    if count is not None and len(solves_flt) > count:
        solves_flt = solves_flt[:count]
        solves_dc = solves_dc[:count]

    bad_solves_count = solves_flt.count(inf)

    # e.g. not set because this is a global mean
    if count is None:
        count = len(solves_dc)

    if operation == "mean":
        if bad_solves_count > 0:
            return Grouping(
                solve_count=count,
                state=State.DNF,
                result=Decimal(0),
                operation=operation,
                solves=solves_dc,
            )
        else:
            return Grouping(
                solve_count=count,
                state=State.SOLVED,
                result=Decimal(mean([s for s in solves_flt if s != inf])),
                solves=solves_dc,
                operation=operation,
            )
    elif operation == "average":
        if bad_solves_count > 1:
            return Grouping(
                solve_count=count,
                state=State.DNF,
                result=Decimal(0),
                operation=operation,
                solves=solves_dc,
            )
        else:
            min_i, max_i = findminmax(solves_flt)
            return Grouping(
                solve_count=count,
                state=State.SOLVED,
                result=Decimal(
                    mean(
                        [s for i, s in enumerate(solves_flt) if i not in {min_i, max_i}]
                    )
                ),
                operation=operation,
                solves=solves_dc,
            )
    elif operation == "global_mean":
        try:
            global_mean = mean([s for s in solves_flt if s != inf])
        except StatisticsError as e:
            return ValueError(str(e) + " - received no valid solves as input")

        return Grouping(
            # dont count DNFs in your global mean 'count'
            solve_count=count - bad_solves_count,
            state=State.SOLVED,
            operation=operation,
            result=Decimal(global_mean),
            solves=solves_dc,
        )
    else:
        raise ValueError(
            f"Unknown operation {operation}, known: 'average', 'mean', 'global_mean'"
        )


def run_operations(
    solves: List[Solve], operation: Operation, counts: List[int]
) -> Dict[int, str]:
    """
    User-facing function to run multiple operations and catch possible errors
    """
    res: Dict[int, str] = {}
    for c in counts:
        gr = grouped(solves, operation=operation, count=c)
        if isinstance(gr, Exception):
            res[c] = "--"
        else:
            res[c] = gr.lhs

    return res


def find_best_group(
    solves: List[Solve], operation: Operation, counts: List[int]
) -> Dict[int, Grouping]:
    res: Dict[int, Grouping] = {}
    solves_flt = solves_to_float(solves)
    for c in counts:
        try:
            res[c] = min(
                (
                    g
                    for g in (
                        grouped(
                            solves_dc=solves[i:],
                            operation=operation,
                            solves_flt=solves_flt[i:],
                            count=c,
                        )
                        for i in range(len(solves) - c + 1)
                    )
                    if not isinstance(g, Exception) and g.state == State.SOLVED
                ),
                key=lambda gr: gr.result,
            )
        except ValueError as e:
            if str(e) != "min() arg is an empty sequence":
                raise e
    return res


def find_best(solves: List[Solve]) -> Solve:
    if len(solves) == 0:
        raise ValueError("Tried to find best solve on empty list")
    min_i = min(list(enumerate(solves_to_float(solves))), key=lambda o: o[1])[0]
    return solves[min_i]



from enum import Enum
from datetime import datetime
from dataclasses import dataclass
from typing import Literal, Optional
from decimal import Decimal

from .timeformat import format_decimal

Operation = Literal["average", "mean", "global_mean"]


class State(Enum):
    SOLVED = "Solved"
    DNS = "DNS"
    DNF = "DNF"


@dataclass(frozen=True)
class Solve:
    """
    The scramble_history 'merged' Solve model
    """

    __slots__ = [
        "puzzle",
        "event_code",
        "event_description",
        "state",
        "scramble",
        "comment",
        "time",
        "penalty",
        "when",
        "full_time",
    ]

    # cstimer: scramble code/manual edit
    # twistytimer: puzzle
    # e.g. 333, 444, 222, pyra, skewb, megaminx
    puzzle: str

    # cstimer scramble code
    # twistytimer category/manually edit
    # What this is: e.g. OH, BLD, LSE, F2L
    event_code: str

    # cstimer CSTimerScramble.name
    # twistytimer category/manually edit
    event_description: str

    # if the cube is solved or not
    state: State
    # standard user-facing stuff here
    scramble: str
    comment: Optional[str]
    time: Decimal
    penalty: Decimal

    # penalty + time
    full_time: Decimal
    when: datetime

    def describe(self) -> str:
        if self.state == State.SOLVED:
            return format_decimal(self.full_time)
        elif self.state == State.DNF:
            return "DNF"
        else:
            return "DNS"



"""
solves from different sources have to be manually mapped/confirmed
from source combinations (tagged with the source name)
to a shared categorization system
"""

import json
import warnings
from pprint import pprint
from pathlib import Path
from typing import NamedTuple, Optional, Any, Dict, List, Iterator, TYPE_CHECKING

if TYPE_CHECKING:
    from prompt_toolkit.completion import FuzzyWordCompleter


from .models import Solve


class SourceMap(NamedTuple):
    # the name of the class, e.g. scramble_history.cstimer, scramble_history.twistytimer
    source_class_name: str

    # these are the fields on the source solve/session
    # that have to match for this transform to be applied
    # as an example using twistytimer:
    # {
    #     "puzzle": "333",
    #     "category": "Normal"
    # }
    #
    # since the 'category'/session name is just a user defined string, user has
    # to be prompted to make sure multiple sources can be merged
    source_fields_match: Dict[str, Any]

    # the normalized fields to use if source_fields_match match
    transformed_puzzle: str
    transformed_event_code: str
    transformed_event_description: str


class SourceMerger:
    def __init__(self, sourcemap_file: Path) -> None:
        self.sourcemap_file = sourcemap_file
        self.sourcemap: List[SourceMap] = []
        self.load()

    def load(self) -> None:
        if self.sourcemap_file.exists():
            self.sourcemap = self.sourcemap_loads(self.sourcemap_file.read_text())

    def dump(self) -> None:
        self.sourcemap_file.write_text(self.sourcemap_dumps(self.sourcemap))

    @staticmethod
    def sourcemap_loads(json_data: str) -> List[SourceMap]:
        data = json.loads(json_data)
        assert isinstance(data, list)
        return [SourceMap(**kw) for kw in data]

    @staticmethod
    def sourcemap_dumps(sm: List[SourceMap]) -> str:
        return json.dumps([d._asdict() for d in sm], indent=4)

    @classmethod
    def _select_keys(cls, data: Any) -> List[str]:

        # define inline so this is only imported when its needed to add new solves
        from prompt_toolkit import prompt
        from prompt_toolkit.document import Document
        from prompt_toolkit.validation import Validator, ValidationError

        class HasattrValidator(Validator):
            def __init__(self, check_obj: Any) -> None:
                super().__init__()
                self.check_obj = check_obj

            def validate(self, document: Document) -> None:
                text = document.text

                for key in text.strip().split():
                    if not hasattr(self.check_obj, key):
                        raise ValidationError(
                            message=f"Could not find key {key} on {self.check_obj}"
                        )

        validator = HasattrValidator(data)
        text = prompt(
            "Which keys (e.g. puzzle, name) should be used to identify solves like this\nIf needed, enter multiple keys, separated by spaces, e.g. 'scramble_code name': ",
            validator=validator,
        )
        return text.strip().split()

    def _create_validator(
        self, key: str, defaults: Dict[str, Any]
    ) -> 'FuzzyWordCompleter':
        tokens = [getattr(s, key) for s in self.sourcemap]
        if defaults.get(key):
            tokens.append(defaults[key])
        return FuzzyWordCompleter(list(set(tokens)))

    def prompt_for_transform(self, data: Any) -> SourceMap:
        from prompt_toolkit import prompt

        pprint(data)
        keys = self._select_keys(data)
        source_fields_match = {k: getattr(data, k) for k in keys}

        defaults = getattr(data, "_prompt_defaults", {})

        transformed_puzzle = prompt(
            "Transformed value for 'puzzle' (e.g. 333, 222, 444, pyra, skewb): ",
            completer=self._create_validator("transformed_puzzle", defaults),
        )

        transformed_event_code = prompt(
            "Transformed value for 'event_code' (e.g. WCA ('normal solving'), LSE, F2L, BLD, OH): ",
            completer=self._create_validator("transformed_event_code", defaults),
        )

        transformed_event_description = prompt(
            "Transformed value for 'event_description (e.g. '3x3 CFOP', 'OH Roux', 'Cross Practice'): ",
            completer=self._create_validator("transformed_event_description", defaults),
        )

        sm = SourceMap(
            source_class_name=self._qualclassname(data),
            source_fields_match=source_fields_match,
            transformed_puzzle=transformed_puzzle,
            transformed_event_code=transformed_event_code,
            transformed_event_description=transformed_event_description,
        )
        self.sourcemap.append(sm)
        self.dump()
        return sm

    @staticmethod
    def _qualclassname(solve: Any) -> str:
        return f"{solve.__module__}.{solve.__class__.__name__}"

    def match_sourcemap(self, solve: Any) -> Optional[SourceMap]:
        for s in self.sourcemap:
            if s.source_class_name != self._qualclassname(solve):
                continue
            if all(
                hasattr(solve, k) and getattr(solve, k) == v
                for k, v in s.source_fields_match.items()
            ):
                return s
        return None

    def match_or_prompt(self, solve: Any) -> SourceMap:
        return self.match_sourcemap(solve) or self.prompt_for_transform(solve)

    def transform(self, solve: Any, sourcemap: Optional[SourceMap] = None) -> Solve:
        if sourcemap is None:
            sourcemap = self.match_or_prompt(solve)
        # each source (cstimer or twistytimer) returns a dict
        # which has the rest of the info for each solve
        transformed_data = solve._transform_map()
        return Solve(
            puzzle=sourcemap.transformed_puzzle,
            event_code=sourcemap.transformed_event_code,
            event_description=sourcemap.transformed_event_description,
            **transformed_data,
        )


from .config import KNOWN_PARSERS, ConfigPaths


def merge(sourcemap_file: Path, conf: ConfigPaths) -> Iterator[Solve]:
    from .source_merger import SourceMerger
    from .cstimer import parse_files as cstimer_merge
    from .twistytimer import parse_files as twistytimer_merge

    merger = SourceMerger(sourcemap_file)

    for flag, grouped_files in conf.items():
        assert flag in KNOWN_PARSERS
        mergefunc = cstimer_merge if flag == "cstimer" else twistytimer_merge
        slv = list(mergefunc(grouped_files))
        if len(slv) == 0:
            warnings.warn(
                f"Did not parse any solves from {flag} {grouped_files}, double check to make sure inputs are correct"
            )
        yield from map(merger.transform, slv)



import shutil
import tempfile
import zipfile
import csv
from typing import (
    Dict,
    Optional,
    List,
    cast,
    Iterator,
    Set,
    Tuple,
    NamedTuple,
    TypeVar,
    Type,
)
from dataclasses import dataclass
from collections import defaultdict
from pathlib import Path
from functools import lru_cache

import requests
import platformdirs

from .log import logger


def cachedir() -> Path:
    return Path(platformdirs.user_cache_dir("wca_export"))


class ExportDownloader:
    def __init__(self) -> None:
        self.cache_dir = cachedir()
        if not self.cache_dir.exists():
            self.cache_dir.mkdir(parents=True)
        self.cache_tsv_dir = self.cache_dir / "tsv"
        self.database_name = "Scores"
        self.export_data_url = (
            "https://www.worldcubeassociation.org/api/v0/export/public"
        )

    @lru_cache(maxsize=1)
    def export_links(self) -> Dict[str, str]:
        req = requests.get(self.export_data_url)
        req.raise_for_status()
        data = req.json()
        assert isinstance(data, dict)
        return cast(Dict[str, str], data)

    @property
    def export_date_path(self) -> Path:
        return self.cache_dir / "export_date.txt"

    def export_date(self) -> Optional[str]:
        if self.export_date_path.exists():
            return self.export_date_path.read_text().strip()
        else:
            return None

    def update_date(self) -> None:
        self.export_date_path.write_text(self.export_links()["export_date"].strip())

    def export_out_of_date(self) -> bool:
        exp_date = self.export_date()
        if exp_date is None:
            return True
        current_date = self.export_links()["export_date"].strip()
        if current_date == exp_date:
            return False
        return True

    def download_export(self) -> None:
        tsv_url = self.export_links()["tsv_url"]
        assert "WCA_export" in tsv_url
        with tempfile.TemporaryDirectory() as td:
            ptd = Path(td)
            assert ptd.exists()
            logger.info("Downloading TSV export...")
            r = requests.get(tsv_url, stream=True)
            assert (
                r.status_code == 200
            ), "Failed to create connection to download TSV export"
            write_to = ptd / "export.zip"
            with open(write_to, "wb") as f:
                for chunk in r:
                    f.write(chunk)

            zip_extract_to = ptd / "archive"

            with zipfile.ZipFile(write_to, "r") as zip_r:
                zip_r.extractall(str(zip_extract_to))

            shutil.copytree(zip_extract_to, self.cache_tsv_dir, dirs_exist_ok=True)
            logger.info(f"Saved TSV export to {self.cache_tsv_dir}")

    def download_if_out_of_date(self) -> None:
        if self.export_out_of_date():
            self.download_export()
            self.update_date()
        else:
            logger.info("Export is already up to date")


TSV = List[str]


def _extract_records(wca_user_id: str, results_file: str) -> Iterator[TSV]:
    with open(results_file, "r", newline="") as f:
        reader = csv.reader(f, delimiter="\t")
        next(reader)
        for line in reader:
            if line[7] == wca_user_id:
                yield line


T = TypeVar("T")


# splat a TSV row onto a namedtuple, handling extra fields if missing
def row_to_type(data: List[str], nt: Type[T]) -> T:
    assert hasattr(nt, "_fields")
    fields: Tuple[str] = cast(Tuple[str], getattr(nt, "_fields"))
    assert isinstance(fields, tuple)
    while len(data) < len(fields):
        data.append("")
    assert len(data) == len(
        fields
    ), f"{data} has too many fields, tried to construct {nt.__name__} {fields}"
    return nt(*data)


class WCA_Result(NamedTuple):
    competitionId: str
    eventId: str
    roundTypeId: str
    pos: str
    best: str
    average: str
    personName: str
    personId: str
    personCountryId: str
    formatId: str
    value1: str
    value2: str
    value3: str
    value4: str
    value5: str
    regionalSingleRecord: str
    regionalAverageRecord: str

    @classmethod
    def parse(cls, data: List[str]) -> "WCA_Result":
        return row_to_type(data, cls)


class WCA_Scramble(NamedTuple):
    scrambleId: str
    competitionId: str
    eventId: str
    roundTypeId: str
    groupId: str
    isExtra: str
    scrambleNum: str
    scramble: str

    @classmethod
    def parse(cls, data: List[str]) -> "WCA_Scramble":
        return row_to_type(data, cls)


def _extract_scrambles_for_competitions(
    competitions_file: str, competitions: Set[str]
) -> Iterator[TSV]:
    with open(competitions_file, "r", newline="") as f:
        reader = csv.reader(f, delimiter="\t")
        next(reader)
        for line in reader:
            if line[1] in competitions:
                yield line


def _match_records_and_scrambles(
    records: List[WCA_Result], scrambles: List[WCA_Scramble]
) -> Iterator[Tuple[WCA_Result, List[WCA_Scramble]]]:
    results_scramble_map: Dict[WCA_Result, List[WCA_Scramble]] = defaultdict(list)
    for record in records:
        for scramble in scrambles:
            if (
                record.competitionId == scramble.competitionId
                and record.eventId == scramble.eventId
                and record.roundTypeId == scramble.roundTypeId
            ):
                results_scramble_map[record].append(scramble)
        if len(results_scramble_map) == 0:
            logger.warning(f"Could not find any scrambles for {record}")
    yield from results_scramble_map.items()


class WCA_Competition(NamedTuple):
    id: str
    name: str
    cityName: str
    countryId: str
    information: str
    year: str
    month: str
    day: str
    endMonth: str
    endDay: str
    cancelled: str
    eventSpecs: str
    wcaDelegate: str
    organiser: str
    venue: str
    venueAddress: str
    venueDetails: str
    external_website: str
    cellName: str
    latitude: str
    longitude: str

    @classmethod
    def parse(cls, data: List[str]) -> "WCA_Competition":
        return row_to_type(data, cls)


def _competition_data(competition_file: str, competitions: Set[str]) -> Iterator[TSV]:
    with open(competition_file, "r", newline="") as f:
        reader = csv.reader(f, delimiter="\t")
        for line in reader:
            if line[0] in competitions:
                yield line


@dataclass
class Details:
    competition_data: List[WCA_Competition]
    results_w_scrambles: List[Tuple[WCA_Result, List[WCA_Scramble]]]


def parse_return_all_details(wca_user_id: str) -> Details:
    exp = ExportDownloader()
    src = exp.cache_tsv_dir
    records = [
        WCA_Result.parse(row)
        for row in _extract_records(wca_user_id, str(src / "WCA_export_Results.tsv"))
    ]
    # all competitions user has been to
    competitions = {r.competitionId for r in records}
    # all scrambles from any competition user has been to
    comp_scrambles = [
        WCA_Scramble.parse(scr)
        for scr in _extract_scrambles_for_competitions(
            str(src / "WCA_export_Scrambles.tsv"), competitions
        )
    ]
    # match records/scrambles
    # TODO: need to find the group that user was in to match to their scrambles?
    results_w_scrambles = list(_match_records_and_scrambles(records, comp_scrambles))
    comp_data = [
        WCA_Competition.parse(c)
        for c in _competition_data(
            str(src / "WCA_export_Competitions.tsv"), competitions
        )
    ]

    return Details(competition_data=comp_data, results_w_scrambles=results_w_scrambles)



from decimal import Decimal


def format_decimal(d: Decimal) -> str:
    """Formats time into h:mm:ss.xxx, removing leftmost places if they are zero"""
    minutes, seconds = divmod(d, 60)
    hours, minutes = divmod(minutes, 60)
    if hours > 0:
        return "{:01d}:{:02d}:{:0>6.3f}".format(int(hours), int(minutes), seconds)
    elif minutes > 0:
        return "{:01d}:{:0>6.3f}".format(int(minutes), seconds)
    else:
        return "{:0>5.3f}".format(seconds)



import sys
import json
from decimal import Decimal
from itertools import chain
from typing import Dict, Any, NamedTuple, List, TextIO, Optional, Tuple, Iterator
from datetime import datetime, timezone
from pathlib import Path

from more_itertools import unique_everseen

from .log import logger
from .cstimer_scramble_type import CSTimerScramble, parse_scramble_type
from .models import State


class SessionSolve(NamedTuple):
    scramble: str
    comment: str
    solve_time: Decimal
    penalty: Decimal
    dnf: bool
    when: datetime


class Session(NamedTuple):
    number: int
    name: str
    raw_scramble_type: str
    scramble_type: Optional[CSTimerScramble]
    solves: List[SessionSolve]


class Solve(NamedTuple):
    number: int
    name: str
    raw_scramble_type: str
    scramble_type: Optional[CSTimerScramble]

    scramble: str
    comment: str
    solve_time: Decimal
    penalty: Decimal
    dnf: bool
    when: datetime

    @property
    def _prompt_defaults(self) -> Dict[str, Any]:
        return {
            "transformed_puzzle": self.raw_scramble_type,
            "transformed_event_description": self.scramble_type.name
            if self.scramble_type is not None
            else None,
        }

    def _transform_map(self) -> Dict[str, Any]:
        return dict(
            state=State.DNF if self.dnf else State.SOLVED,
            scramble=self.scramble,
            comment=self.comment,
            time=self.solve_time,
            penalty=self.penalty,
            when=self.when,
            full_time=self.solve_time + self.penalty,
        )


def parse_file(path: Path) -> List[Session]:
    with path.open("r") as f:
        return _parse_blob(f)


def denormalize(sessions: List[Session]) -> Iterator[Solve]:
    for sess in sessions:
        for solve in sess.solves:
            yield Solve(
                number=sess.number,
                name=sess.name,
                raw_scramble_type=sess.raw_scramble_type,
                scramble_type=sess.scramble_type,
                scramble=solve.scramble,
                comment=solve.comment,
                solve_time=solve.solve_time,
                penalty=solve.penalty,
                dnf=solve.dnf,
                when=solve.when,
            )


def _parse_blob(f: TextIO) -> List[Session]:
    data = json.loads(f.read())
    props: Dict[str, Any] = data["properties"]
    session_raw: str = props["sessionData"]
    assert isinstance(
        session_raw, str
    ), "Fatal error parsing sessions, expected sessionData to be string"
    session_info = json.loads(session_raw)

    sessions: List[Session] = []

    # parse each session
    for session_number, session_val in session_info.items():
        # e.g. for session_number '1' -> key in top-level data
        # is "session1"
        data_key = f"session{session_number}"
        if data_key not in data:
            logger.debug(
                f"Expected session key '{data_key}' in data, ignoring session '{session_val}'"
            )
            continue

        session_name = session_val["name"]

        options = session_val.get("opt", {})
        # default to WCA 333 scramble if unset
        scramble_code = options.get("scrType", "333")
        raw_scrambles = data[data_key]
        scrambles = map(_parse_scramble, raw_scrambles)

        scramble_type: Optional[CSTimerScramble] = None
        try:
            scramble_type = parse_scramble_type(scramble_code)
        except KeyError:
            pass

        sessions.append(
            Session(
                number=int(session_number),
                name=session_name,
                raw_scramble_type=scramble_code,
                scramble_type=scramble_type,
                solves=[s for s in scrambles if s is not None],
            )
        )

    return sessions


RawScramble = Tuple[Tuple[int, int], str, str, int]


def _parse_scramble(raw: RawScramble) -> Optional[SessionSolve]:
    try:
        [[penalty, solve_time], scramble, comment, timestamp] = raw
        is_dnf = penalty == -1

        # if this was a DNF (did not finish), we should remove the penalty (is marked as '-1')
        if is_dnf:
            penalty = 0

        return SessionSolve(
            scramble=scramble.strip(),
            comment=comment,
            solve_time=Decimal(solve_time) / 1000,
            penalty=Decimal(penalty) / 1000,
            dnf=is_dnf,
            when=datetime.fromtimestamp(timestamp, tz=timezone.utc),
        )
    except ValueError as e:
        print(f"Could not parse raw scramble info for {raw}: {e}", file=sys.stderr)
        return None


def parse_files(paths: List[Path]) -> Iterator[Solve]:
    yield from unique_everseen(chain(*(denormalize(parse_file(p)) for p in paths)))



from typing import Union, List, NamedTuple, Literal, Tuple

from .average_parser import parse_operation_code
from .group_operations import grouped, find_best
from .error import unwrap
from .models import Operation, Solve


class Filter(NamedTuple):
    attr: str
    value: str


class Average(NamedTuple):
    operation: Operation
    count_: int


class Drop(NamedTuple):
    count_: int


class Limit(NamedTuple):
    count_: int


Commands = Literal["dump", "best"]

QueryPart = Union[Filter, Average, Commands, Drop, Limit]

Query = List[QueryPart]


def parse_query(inputs: Union[str, List[str]]) -> Query:
    raw_tokens = []
    if isinstance(inputs, str):
        raw_tokens.append(inputs)
    else:
        raw_tokens.extend(inputs)

    tokens = []
    for rt in raw_tokens:
        if "___" in rt:
            tokens.extend(rt.split("___"))
        else:
            tokens.append(rt)

    assert isinstance(inputs, list)
    parsed: Query = []
    for token in tokens:
        if "==" in token:
            solve_attr, value = token.split("==", maxsplit=1)
            parsed.append(Filter(solve_attr, value))
            continue

        try:
            op, count = parse_operation_code(token)
            parsed.append(Average(op, count))
            continue
        except ValueError:
            pass

        if token.lower() == "dump":
            parsed.append("dump")
        elif token.lower() == "best":
            parsed.append("best")
        elif token.lower().startswith("drop:"):
            parsed.append(Drop(int(token.split("drop:", maxsplit=1)[-1])))
        elif token.lower().startswith("limit:"):
            parsed.append(Limit(int(token.split("limit:", maxsplit=1)[-1])))

        else:
            raise ValueError(f"Query: not sure how to parse token '{token}'")

    return parsed


QueryRet = Union[Tuple[str, ...], List[Solve]]


def run_query(solves: List[Solve], *, query: Query) -> QueryRet:
    returns: List[str] = []
    for qr in query:
        if isinstance(qr, Filter):
            if len(solves) == 0:
                continue
            assert hasattr(
                solves[0], qr.attr
            ), f"could not find attribute {qr} on {solves[0]}"
            solves = list(filter(lambda solv: getattr(solv, qr.attr) == qr.value, solves))  # type: ignore[arg-type]
        elif isinstance(qr, Average):
            g = unwrap(grouped(solves, operation=qr.operation, count=qr.count_))
            returns.append(g.describe())
        elif isinstance(qr, Drop):
            solves = solves[qr.count_ :]
        elif isinstance(qr, Limit):
            solves = solves[: qr.count_]
        else:
            if qr == "best":
                returns.append(find_best(solves).describe())
            else:
                assert qr == "dump", str(qr)
                returns.append("\n".join([s.describe() for s in solves]))

    if len(returns) == 0:
        return solves

    return tuple(returns)



import os
import logging
from typing import Optional

from logzero import setup_logger  # type: ignore[import]

DEFAULT_LEVEL = logging.INFO


def setup(level: Optional[int] = None) -> logging.Logger:
    chosen_level = level or int(os.environ.get("SCRAMBLE_HISTORY_LOGS", DEFAULT_LEVEL))
    lgr: logging.Logger = setup_logger(name=__package__, level=chosen_level)
    return lgr


logger = setup()



import io
import warnings
from pathlib import Path
from glob import glob as do_glob
from collections import defaultdict
from typing import Iterable, List, Union, Optional, Dict, Mapping

PathIsh = Union[Path, str]
Paths = Union[PathIsh, List[PathIsh]]


DEFAULT_GLOB = "*"


# Modified from: https://github.com/karlicoss/HPI
def get_files(
    pp: Paths,
    glob: str = DEFAULT_GLOB,
) -> List[Path]:
    sources: List[Path]
    if isinstance(pp, Path):
        sources = [pp]
    elif isinstance(pp, str):
        sources = [Path(pp)]
    else:
        assert isinstance(
            pp, list
        ), f"Input should either be a Path, string or list of paths, found {type(pp)}"
        sources = [Path(p) for p in pp]

    paths: List[Path] = []
    for src in sources:
        if src.parts[0] == "~":
            src = src.expanduser()
        # note: glob handled first, because e.g. on Windows asterisk makes is_dir unhappy
        gs = str(src)
        if "*" in gs:
            if glob != DEFAULT_GLOB:
                warnings.warn(
                    f"treating {gs} as glob path. Explicit glob={glob} argument is ignored!"
                )
            paths.extend(map(Path, do_glob(gs)))
        elif src.is_dir():
            gp: Iterable[Path] = src.glob(glob)
            paths.extend(gp)
        else:
            if not src.is_file():
                # todo not sure, might be race condition?
                raise RuntimeError(f"Expected '{src}' to exist")
            paths.append(src)

    return paths


ConfigPaths = Dict[str, List[Path]]


def parse_config_file(file: Path) -> Dict[str, List[Path]]:
    return parse_config(file.read_text())


def parse_config(data: str) -> Dict[str, List[Path]]:
    import yaml

    buf = io.StringIO(data)
    loaded = yaml.load(buf, Loader=yaml.FullLoader)
    res: ConfigPaths = {}
    for k, v in loaded.items():
        res[k] = get_files(v)
    return res


KNOWN_PARSERS = {"cstimer", "twistytimer"}


# used in the CLI to parse unprocessed arguments
def group_args_by_options(args: List[str]) -> ConfigPaths:
    parser: Optional[str] = None
    parsed: Mapping[str, List[Path]] = defaultdict(list)
    for p in args:
        if parser is None or p.startswith("--"):
            parser = p.strip().strip("-")
        else:
            pp = Path(p)
            if not pp.exists():
                raise ValueError(f"Filepath '{pp}' does not exist")
            parsed[parser].append(pp)
    return dict(parsed)


def check_config(conf: ConfigPaths) -> None:
    for k in conf:
        if k not in KNOWN_PARSERS:
            raise ValueError(f"Could not find {k} in known sources: {KNOWN_PARSERS}")



import warnings
from typing import List, Union, Optional, Tuple
from decimal import Decimal

from .models import Operation


def _is_float_like(s: str) -> bool:
    try:
        float(s)
        return True
    except ValueError:
        return False


def parse_average(average_str: str) -> List[Union[Decimal, str]]:
    """
    Parses times that look like:
    3x3: 22.03 = (DNF), 25.14, 21.69, 19.26, (25.63)
    22.03 = (DNF), 25.14, 21.69, 19.26, (25.63)
    (DNF), 25.14, 21.69, 19.26, (25.63)

    into individual times, like:
    [
        "DNF",
        25.14,
        21.69,
        19.26,
        25.63
    ]

    If it cannot parse a token/time, it leaves it as a string
    """
    import pytimeparse  # type: ignore[import]

    if "=" in average_str:
        average_str = average_str.split("=", maxsplit=1)[-1]
    average_str = average_str.strip()
    # split by commas or spaces
    raw_solves = []
    if "," in average_str:
        raw_solves = average_str.split(",")
    elif " " in average_str:
        raw_solves = average_str.split()
    else:
        raw_solves = [average_str]
    # remove parens
    raw_solves = [s.strip().lstrip("(").rstrip(")") for s in raw_solves]
    solves: List[Union[Decimal, str]] = []
    for s in raw_solves:
        if s.lower() in ("dns", "dnf"):
            solves.append(s.upper())
        else:
            if _is_float_like(s):
                solves.append(Decimal(s))
                continue
            td: Optional[Union[float, int]] = pytimeparse.parse(s)
            if td is None:
                solves.append(s)
                warnings.warn(f"Warning: Not sure how to parse token {s}")
            else:
                solves.append(Decimal(td))
    return solves


def parse_operation_code(op: str) -> Tuple[Operation, int]:
    op = op.lower().lower()
    if op.startswith("ao"):
        return "average", int(op[2:])
    elif op.startswith("mo"):
        return "mean", int(op[2:])
    raise ValueError(f"{op} does not start with 'Ao' (average of) or 'Mo' (mean)")



from pathlib import Path
from setuptools import setup, find_packages

long_description = Path("README.md").read_text()
reqs = Path("requirements.txt").read_text().strip().splitlines()

pkg = "scramble_history"
setup(
    name=pkg,
    version="0.1.7",
    url="https://github.com/seanbreckenridge/scramble-history",
    author="Sean Breckenridge",
    author_email="seanbrecke@gmail.com",
    description=("""parses scramble history from cstimer.net and other sources"""),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(include=[pkg]),
    install_requires=reqs,
    package_data={pkg: ["py.typed"]},
    zip_safe=False,
    keywords="cubing data",
    python_requires=">=3.8",
    entry_points={
        "console_scripts": ["scramble_history = scramble_history.__main__:main"]
    },
    extras_require={
        "optional": ["orjson"],
        "testing": [
            "pytest",
            "mypy",
            "flake8",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import os
import itertools
from pathlib import Path
from typing import Any, List, Dict, Sequence, Optional
from decimal import Decimal

import click

from .config import (
    parse_config_file,
    ConfigPaths,
    group_args_by_options,
    check_config,
)
from .query import parse_query, run_query, Query


def _default(o: Any) -> Any:
    # orjson doesn't serialize namedtuples to avoid serializing
    # them as tuples (arrays), since they're technically a subclass
    if isinstance(o, Decimal):
        return str(o)
    if hasattr(o, "_asdict"):
        return o._asdict()
    raise TypeError(f"Could not serialize object of type {type(o).__name__}")


def _serialize(data: Any) -> str:
    import orjson  # type: ignore[import]

    bdata: bytes = orjson.dumps(
        data,
        option=orjson.OPT_NON_STR_KEYS,
        default=_default,
    )
    return bdata.decode("utf-8")


JSON = click.option(
    "-j", "--json", "_json", is_flag=True, default=False, help="print data as JSON"
)


@click.group(context_settings={"max_content_width": 110})
def main() -> None:
    """
    parses your rubiks cube scramble history
    """
    pass


@main.group()
def export() -> None:
    """
    Export data from a website
    """


@export.group(name="wca")
def _wca_export() -> None:
    """
    Data from the worldcubeassosiation.org website
    """


@_wca_export.command()
def update() -> None:
    """
    Download/update the local TSV data if its out of date
    """
    from .wca_export import ExportDownloader

    exp = ExportDownloader()
    exp.download_if_out_of_date()


@_wca_export.command()
@click.option(
    "-u", "--wca-user-id", type=str, help="WCA ID to extract results for", required=True
)
@JSON
def extract(_json: bool, wca_user_id: str) -> None:
    """
    Extract details from the local TSV data (must call update first)
    """
    from .wca_export import parse_return_all_details

    details = parse_return_all_details(wca_user_id)
    if _json:
        click.echo(_serialize(details))
    else:
        import IPython  # type: ignore[import]

        header = f"Use {click.style('details', fg='green')} to review TSV data"
        IPython.embed(header=header)


@main.group()
def parse() -> None:
    """
    Parse the output of some file/directory
    """
    pass


@parse.command(short_help="parse cstimer.net export file")
@JSON
@click.argument(
    "CSTIMER_FILE",
    required=True,
    type=click.Path(exists=True, path_type=Path),
)
def cstimer(_json: bool, cstimer_file: Path) -> None:
    """
    Expects the cstimer.net export file as input
    """
    from .cstimer import parse_file

    sess = parse_file(cstimer_file)
    if _json:
        click.echo(_serialize(sess))
    else:
        import IPython  # type: ignore[import]

        header = f"Use {click.style('sess', fg='green')} to review session data"
        IPython.embed(header=header)


@parse.command(short_help="parse twistytimer export file")
@click.argument(
    "TWISTYTIMER_FILE",
    required=True,
    type=click.Path(exists=True, path_type=Path),
)
@JSON
def twistytimer(_json: bool, twistytimer_file: Path) -> None:
    """
    Expects the twistytimer export file as input

    This works for both the cubers.io and twistytimer export
    """
    from .twistytimer import parse_file

    solves = list(parse_file(twistytimer_file))
    if _json:
        click.echo(_serialize(solves))
    else:
        import IPython  # type: ignore[import]

        header = f"Use {click.style('solves', fg='green')} to review your solves"
        IPython.embed(header=header)


config_dir = Path(os.environ.get("XDG_CONFIG_DIR", Path.home() / ".config"))

# this needs to be a global path that user cant modify in click option
# so _parse_merge_inputs can access it
config_file = Path(
    os.environ.get("SCRAMBLE_HISTORY_CONFIG", config_dir / "scramble_history.yaml")
)


def _parse_merge_inputs(
    ctx: click.Context, param: click.Argument, value: Sequence[str]
) -> ConfigPaths:
    conf = {}
    if config_file.exists():
        conf = parse_config_file(config_file)

    if len(value) < 1 and len(conf) == 0:
        raise click.BadArgumentUsage(
            f"Must supply options/datafiles as input or create a config file at {config_file}"
        )
    try:
        # merge any options into dictionary if supplied in addition to config file
        for k, v in group_args_by_options(list(value)).items():
            conf[k] = conf.get(k, []) + v
    except ValueError as e:
        raise click.BadArgumentUsage(str(e))

    check_config(conf)
    return conf


def _parse_query(
    ctx: click.Context, param: click.Argument, value: Sequence[str],
) -> Query:
    return parse_query(list(value))


def banner() -> None:
    click.echo("===================")


@main.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
        max_content_width=110,
    )
)
@click.option(
    "-s",
    "--sourcemap-file",
    help="Data file which saves choices on how to map solves from different sources",
    default=config_dir / "scramble_history_sourcemap.json",
    show_default=True,
    type=click.Path(dir_okay=False, path_type=Path),
)
@click.option(
    "-a",
    "--action",
    type=click.Choice(["json", "repl", "stats"]),
    help="what to do with merged solves",
    default="repl",
    show_default=True,
)
@click.option(
    "-C",
    "--check",
    help="Dont print/interact, just check that all solves are transformed properly",
    is_flag=True,
    default=False,
)
@click.option(
    "-g",
    "--group-by",
    type=click.Choice(["puzzle", "event_code", "event_description"]),
    help="Group parsed results by key",
    default=None,
)
@click.option(
    "-q",
    "--query",
    type=click.UNPROCESSED,
    default=(),
    multiple=True,
    callback=_parse_query,
    help="Solves to filter to, or actions to run",
)
@click.option(
    "-s",
    "--sort-by",
    type=click.Choice(["when"]),
    default="when",
    help="Sort the resulting solves",
)
@click.option(
    "-r",
    "--reverse/--no-reverse",
    is_flag=True,
    default=True,
    help="Reverse the sort for --sort-by. --reverse is the default",
)
@click.argument(
    "DATAFILES",
    type=click.UNPROCESSED,
    callback=_parse_merge_inputs,
    default=(),
)
def merge(
    sourcemap_file: Path,
    action: str,
    check: bool,
    sort_by: Optional[str],
    reverse: bool,
    query: Optional[Query],
    group_by: Optional[str],
    datafiles: Dict[str, List[Path]],
) -> None:
    """
    merge solves from different data sources together
    """
    from .source_merger import merge as merge_solves

    solves = list(merge_solves(sourcemap_file=sourcemap_file, conf=datafiles))

    if check:
        return

    if sort_by is not None:
        if sort_by == "when":
            solves.sort(key=lambda s: s.when, reverse=reverse)

    if query:
        data = run_query(solves, query=query)
        # if these were not just a Filter and this modified
        # the shape/ran something, we should show that
        if isinstance(data, tuple):
            for resp in data:
                print(resp)
            return
        else:
            # we just filtered, so set the solves to what the query returned
            solves = data

    res: Any = solves
    if group_by is not None or action == "stats" and isinstance(res, list):
        if group_by is None:
            click.echo(
                "Passed 'stats' with no '--group_by', grouping by 'event_description'",
                err=True,
            )
            group_by = "event_description"
        key = str(group_by)
        assert hasattr(
            solves[0], key
        ), f"Error: could not find attribute {key} on {solves[0]}"
        solves.sort(key=lambda s: getattr(s, key))  # type: ignore[no-any-return]
        res = {
            k: list(g)
            for k, g in itertools.groupby(solves, key=lambda s: getattr(s, key))  # type: ignore[no-any-return]
        }

    if action == "json":
        click.echo(_serialize(res))
    elif action == "repl":
        import IPython  # type: ignore[import]

        header = f"Use {click.style('res', fg='green')} to review"
        IPython.embed(header=header)
    else:
        from .group_operations import run_operations, grouped, find_best_group, operation_code, find_best
        from .timeformat import format_decimal
        from tabulate import tabulate

        for group_name, group_solves in res.items():
            group_solves.sort(key=lambda s: s.when, reverse=reverse)
            banner()
            click.echo(group_name)
            banner()
            recent_ao5 = grouped(group_solves, count=5, operation="average")
            desc = (
                "--"
                if isinstance(recent_ao5, Exception)
                else recent_ao5.describe_average()
            )
            click.echo(f"Best => {find_best(group_solves).describe()}")
            click.echo(f"Most recent Ao5 => {desc}")
            click.echo(f"Solve Count => {len(group_solves)}")
            stat_data = run_operations(
                group_solves, operation="average", counts=[5, 12, 50, 100]
            )
            best = find_best_group(group_solves, operation="average", counts=[5, 12, 50, 100])
            click.echo()
            click.echo(
                tabulate(
                    [
                        [
                            operation_code("average", count_, count_),
                            stat_data[count_],
                            format_decimal(best[count_].result)
                            if count_ in best
                            else "--",
                        ]
                        for count_ in stat_data.keys()
                    ],
                    headers=(group_name, "Current", "Best"),
                )
            )


if __name__ == "__main__":
    main(prog_name="scramble_history")



from typing import Union, TypeVar

T = TypeVar("T")
E = TypeVar("E", bound=Exception)

ResT = Union[T, E]

Res = ResT[T, Exception]


def unwrap(res: Res[T]) -> T:
    if isinstance(res, Exception):
        raise res
    else:
        return res
